/*
 * Copyright (c) 2004, Stefan Walter
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 *     * Redistributions of source code must retain the above 
 *       copyright notice, this list of conditions and the 
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the 
 *       above copyright notice, this list of conditions and 
 *       the following disclaimer in the documentation and/or 
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be 
 *       used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGE.
 * 
 *
 * CONTRIBUTORS
 *  Stef Walter <stef@memberwebs.com>
 *
 */ 
 
#ifndef __RTFPARSER_H__
#define __RTFPARSER_H__

#include "usuals.h"
#include <stack>
#include <stdio.h>

class RtfHandler;

/*
 * RtfParser
 * 
 * A class that parses the RTF into it's tags and groups etc... It feeds its
 * parsed data into into a handler interface (see below) for processing.
 * 
 * Performs some basic conversion and sanity checking (unicode chars etc...) 
 * as well.
 */ 
class RtfParser
{
public:
	RtfParser();
	virtual ~RtfParser();

	bool parse(string fileName);
	bool parse(FILE* file);

	void setHandler(RtfHandler* handler)
		{ m_handler = handler; }
	string getParseErrors() const
		{ return m_parseErrors; }
	int getDepth() const
		{ return m_depth; }
	void setHexParse(bool parse)
		{ m_parseHex = parse; }
	void setUnicode(bool unicode)
		{ m_parseUnicode = unicode; }

protected:
	RtfHandler* m_handler;         // The current handler
	int m_depth;                   // To keep track of group depth 
	bool m_parseHex;               // Whether to parse hex chars or not
	string m_parseErrors;          // A list of all the RTF parsing errors

    // Unicode char handling
	bool m_parseUnicode;               
	typedef std::stack<int> StackInt;
	StackInt m_uniEatStack;
	int m_uniEat;

    FILE* m_file;                   // The file we're currently parsing
    string m_dataBuffer;            // The block of data we're caching to send
    
private:

    // Parse helpers
	bool parseControlWord();
	bool parseHexChar(int num, bool ansi);
      
    // Convenience functions for calling the handler
	void sendControlWord(const string& cw, int flags, int param);
	void sendData(wchar_t ch);
    void sendData(const wstring& data);
	void flushData();
};

/*
 * RtfHandler
 * 
 * An interface called by RtfParser with tags and groups etc... parsed from
 * an RTF file. 
 */  
class RtfHandler  
{
public:
    virtual ~RtfHandler() {};

    // Called at the beginning of the document
    virtual void startDocument(RtfParser* reader) = 0;
    
    // Called at the end of the document 
    virtual void endDocument() = 0;
    
    // Called when an RTF control word is hit. Flags below.
    // If control word has no param then param is -1
    virtual void controlWord(const string& cw, int flags, int param) = 0;
    
    // Called when an RTF group opened
    virtual void groupStart() = 0; 
    
    // Called when an RTF group is closed
    virtual void groupEnd() = 0;
    
    // A block of character data encountered
    virtual void charData(const string& data) = 0;

    // Flags for controlWord
    enum
    {
        kAsterisk = 0x00000001,
        kHasParam = 0x00000002,
        kIsEncoded = 0x00000004
    };
};

#endif // __RTFPARSER_H__
